"""
Network
=======

#. :class:`.Network`

Class for defining a :mod:`networkx` graph from a molecule.

"""

import logging
import networkx as nx

logger = logging.getLogger(__name__)


class PositionedAtom:
    """
    A container for stk.Atom and a position.

    """

    def __init__(self, atom, position):
        """
        Initialize a :class:`PositionedAtom`.

        Parameters
        ----------
        atom : :class:`stk.Atom`
            The atom.

        position : : :class:`tuple` of :class:`float`
            The position of the atom.

        """

        self._atom = atom
        self._position = position

    def get_atom(self):
        return self._atom

    def get_atomic_number(self):
        return self._atom.get_atomic_number()

    def get_charge(self):
        return self._atom.get_charge()

    def get_id(self):
        return self._atom.get_id()

    def get_position(self):
        return self._position

    def _with_id(self, id):
        """
        Modify the atom.

        """

        self._atom = self._atom.with_id(id)
        return self

    def with_id(self, id):
        """
        Get a clone but with a different id.

        Returns
        -------
        :class:`.Atom`
            A clone with a new id. Has the same type as the original
            atom.

        """

        return self.clone()._with_id(id)

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.Atom`
            The clone. It has the same type as the original atom.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._atom = self._atom
        clone._position = self._position
        return clone

    def __repr__(self):
        return (
            f'{self._atom.__class__.__name__}({self._atom.get_id()})'
        )

    def __str__(self):
        return repr(self)


class Network:
    """
    Definition of a network of an stk.Molecule.

    """

    def __init__(self, graph):
        """
        Initialize a Network from a networkx.graph.

        """

        self._graph = graph

    @classmethod
    def init_from_molecule(cls, molecule):
        """
        Initialize a Network from a stk.Molecule.

        """

        g = nx.Graph()

        pos_mat = molecule.get_position_matrix()
        for atom in molecule.get_atoms():
            pos = tuple(float(i) for i in pos_mat[atom.get_id()])
            pa = PositionedAtom(atom, pos)
            g.add_node(pa)

        # Define edges.
        for bond in molecule.get_bonds():
            n1, n2 = [
                i for i in g.nodes
                if i.get_id() in (
                    bond.get_atom1().get_id(),
                    bond.get_atom2().get_id(),
                )
            ]

            g.add_edge(
                n1, n2,
                order=bond.get_order(),
                periodicity=bond.get_periodicity(),
            )

        return cls(g)

    def get_graph(self):
        """
        Return a networkx.graph.

        """

        return self._graph

    def get_nodes(self):
        """
        Yield nodes of networkx.graph.

        """

        for i in self._graph.nodes:
            yield i

    def clone(self):
        """
        Return a clone.

        """

        clone = self.__class__.__new__(self.__class__)
        Network.__init__(self=clone, graph=self._graph)
        return clone

    def _with_deleted_bonds(self, atom_ids):
        sorted_set = {tuple(sorted(i)) for i in atom_ids}
        to_delete = []
        for edge in self._graph.edges:
            a1id = edge[0].get_id()
            a2id = edge[1].get_id()
            pair = tuple(sorted((a1id, a2id)))
            if pair in sorted_set:
                to_delete.append(edge)

        for id1, id2 in to_delete:
            self._graph.remove_edge(id1, id2)

        return self

    def with_deleted_bonds(self, atom_ids):
        """
        Return a clone with edges between `atom_ids` deleted.

        """

        return self.clone()._with_deleted_bonds(atom_ids)

    def get_connected_components(self):
        """
        Get connected components within full graph.

        Returns
        -------
        :class:`list` of :class:`networkx.graph`
            List of connected components of graph.

        """

        return [
            self._graph.subgraph(c).copy()
            for c in sorted(
                nx.connected_components(self._graph)
            )
        ]

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'n={self._graph.number_of_nodes()}, '
            f'e={self._graph.number_of_edges()})'
        )



import stko
from copy import deepcopy
from .utilities import (
    inequivalent_position_matrices,
    is_equivalent_molecule,
)


def test_optimizer_sequence(
    passing_optimizer, unoptimized_mol
):

    opts = [deepcopy(passing_optimizer) for i in range(10)]
    opt_seq = stko.OptimizerSequence(*opts)
    opt_res = opt_seq.optimize(unoptimized_mol)
    is_equivalent_molecule(opt_res, unoptimized_mol)
    inequivalent_position_matrices(opt_res, unoptimized_mol)


def test_trycatchoptimizer(
    passing_optimizer, failing_optimizer, unoptimized_mol,
):

    opt = stko.TryCatchOptimizer(
        try_optimizer=failing_optimizer,
        catch_optimizer=passing_optimizer,
    )
    opt_res = opt.optimize(unoptimized_mol)
    is_equivalent_molecule(opt_res, unoptimized_mol)
    inequivalent_position_matrices(opt_res, unoptimized_mol)



"""
TorsionInfo
===========

Class defining torsion angle info.

"""

import logging

logger = logging.getLogger(__name__)


class TorsionInfo:
    """
    Holds additional info about ConstructedMoleculeTorsioned torsions.

    """

    def __init__(
        self,
        torsion,
        building_block,
        building_block_id,
        building_block_torsion,
    ):
        """
        Initialize TorsionInfo.

        Parameters
        ----------
        torsion : :class:`.Torsion`
            The torsion about which information is held.

        building_block : :class:`stk.BuildingBlock`
            The building block from which this torsion originates.
            Can be ``None``, if the atoms that make up the torsion did
            not come from a single building block.

        building_block_id : :class:`int`
            A unique id for each :class:`.Molecule` placed during
            the construction of the :class:`.ConstructedMolecule`. As a
            single :class:`.Molecule` can be placed multiple times
            during construction, the `building_block_id` allows
            the user to distinguish between each placement. Can be
            Can be ``None``, if the atoms that make up the torsion did
            not come from a single building block.

        building_block_torsion : :class:`.Torsion`
            The building block torsion from which this torsion
            originates. Can be ``None``, if the atoms that make
            up the torsion did not come from a single building block.

        """

        self._torsion = torsion
        self._building_block = building_block
        self._building_block_id = building_block_id
        self._building_block_torsion = building_block_torsion

    def get_torsion(self):
        """
        Torsion of atoms in constructed molecule.

        """

        return self._torsion

    def get_building_block(self):
        return self._building_block

    def get_building_block_torsion(self):
        """
        Torsion of atoms in building block.

        """

        return self._building_block_torsion

    def get_building_block_id(self):
        return self._building_block_id



"""
Topology Info
=============

#. :class:`.TopologyInfo`

Class containing extracted topology information.

"""

import logging

logger = logging.getLogger(__name__)


class TopologyInfo:
    """
    Extracted information of a topology.

    """

    def __init__(self, centroids, connectivities, edge_pairs):
        """


        """

        self._centroids = centroids
        self._connectivities = connectivities
        self._edge_pairs = edge_pairs

    def get_vertex_positions(self):
        """
        Get the positions of each vertex.

        Returns
        -------
        :class:`dict`
            Vertex ids with their positions.

        """

        return self._centroids

    def get_connectivities(self):
        """
        Get the number of connections of each vertex.

        Returns
        -------
        :class:`dict`
            Vertex ids with their number of connections.

        """

        return self._connectivities

    def get_edge_pairs(self):
        """
        Get the edge pairs.

        Returns
        -------
        :class:`list` of :class:`tuple`
            List of edge pairs.

        """

        return self._edge_pairs

    def write(self, path):
        """
        Writes a mock .pdb with vertex centroids and edges as bonds.

        """

        content = []

        atom_counts = {}
        hetatm = 'HETATM'
        alt_loc = ''
        res_name = 'UNL'
        chain_id = ''
        res_seq = '1'
        i_code = ''
        occupancy = '1.00'
        temp_factor = '0.00'

        # This set will be used by bonds.
        atoms = set()
        for cent in self._centroids:
            atoms.add(cent)
            serial = cent+1
            element = 'P'
            charge = 0
            atom_counts[element] = atom_counts.get(element, 0) + 1
            name = f'{element}{atom_counts[element]}'
            # Make sure the coords are no more than 8 columns wide
            # each.
            x, y, z = self._centroids[cent]

            content.append(
                f'{hetatm:<6}{serial:>5} {name:<4}'
                f'{alt_loc:<1}{res_name:<3} {chain_id:<1}'
                f'{res_seq:>4}{i_code:<1}   '
                f' {x:>7.3f} {y:>7.3f} {z:>7.3f}'
                f'{occupancy:>6}{temp_factor:>6}          '
                f'{element:>2}{charge:>2}\n'
            )

        conect = 'CONECT'
        for edge in self._edge_pairs:
            a1 = edge[0]
            a2 = edge[1]
            if a1 in atoms and a2 in atoms:
                content.append(
                    f'{conect:<6}{a1+1:>5}{a2+1:>5}               \n'
                )

        content.append('END\n')

        with open(path, 'w') as f:
            f.write(''.join(content))

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'<{self.__class__.__name__} at {id(self)}>'





def pytest_addoption(parser):
    parser.addoption('--macromodel_path', default='')


def pytest_generate_tests(metafunc):
    if 'macromodel_path' in metafunc.fixturenames:
        macromodel_path = metafunc.config.getoption('macromodel_path')
        metafunc.parametrize('macromodel_path', [macromodel_path])



"""
Collapser Optimizer
===================

#. :class:`.Collapser`
#. :class:`.CollapserMC`

Optimizer for collapsing enlarged topologies.

"""

import logging
from itertools import combinations
import numpy as np
from collections import defaultdict
from scipy.spatial.distance import pdist
import random
import matplotlib.pyplot as plt
import uuid
import os
import shutil
from stk import PdbWriter

from .optimizers import Optimizer
from ..utilities import get_atom_distance


logger = logging.getLogger(__name__)


class Collapser(Optimizer):
    """
    Collapse stk.ConstructedMolecule to decrease enlarged bonds.

    It is recommended to use the MCHammer version of this code with
    :mod:`MCHammer` [1]_, where a much cleaner version is written.
    The utilities `get_long_bond_ids` will help generate sub units.

    This optimizer aims to bring extended bonds closer together for
    further optimisation.

    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock(
            smiles='O=CC(C=O)C=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        cage1 = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusSix((bb1, bb2)),
        )

        # Perform collapser optimisation.
        optimizer = stko.Collapser(
            output_dir='test_coll',
            step_size=0.05,
            distance_cut=2.0,
            scale_steps=True,
        )
        cage1 = optimizer.optimize(mol=cage1)

    References
    ----------
    .. [1] https://github.com/andrewtarzia/MCHammer

    """

    def __init__(
        self,
        output_dir,
        step_size,
        distance_cut,
        scale_steps=True,
    ):
        """
        Initialize a :class:`Collapser` instance.

        Parameters
        ----------
        output_dir : :class:`str`
            The name of the directory into which files generated during
            the calculation are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        step_size : :class:`float`
            The relative size of the step to take during collapse.

        distance_cut : :class:`float`
            Distance between distinct building blocks to use as
            threshold for halting collapse in Angstrom.

        scale_steps : :class:`bool`, optional
            Whether to scale the step of each distict building block
            by their relative distance from the molecules centroid.
            Defaults to ``True``

        """

        self._output_dir = output_dir
        self._step_size = step_size
        self._distance_cut = distance_cut
        self._scale_steps = scale_steps

    def _get_inter_bb_distance(self, mol):
        """
        Yield The distances between building blocks in mol.

        Ignores H atoms.

        """

        position_matrix = mol.get_position_matrix()

        for atom1, atom2 in combinations(mol.get_atom_infos(), 2):
            chk1 = (
                atom1.get_atom().get_id() != atom2.get_atom().get_id()
            )
            chk2 = (
                atom1.get_atom().get_atomic_number() != 1
                and atom2.get_atom().get_atomic_number() != 1
            )
            chk3 = (
                atom1.get_building_block_id() !=
                atom2.get_building_block_id()
            )
            if chk1 and chk2 and chk3:
                dist = get_atom_distance(
                    position_matrix=position_matrix,
                    atom1_id=atom1.get_atom().get_id(),
                    atom2_id=atom2.get_atom().get_id()
                )
                yield dist

    def _has_short_contacts(self, mol):
        """
        Calculate if there are short contants in mol.

        """

        return any(
            dist < self._distance_cut
            for dist in self._get_inter_bb_distance(mol)
        )

    def _get_new_position_matrix(self, mol, step, vectors, scales):
        """
        Get the position matrix of the mol after translation.

        """

        new_position_matrix = mol.get_position_matrix()
        for atom in mol.get_atom_infos():
            bb_id = atom.get_building_block_id()
            _id = atom.get_atom().get_id()
            pos = mol.get_position_matrix()[_id]
            new_position_matrix[_id] = (
                pos - step*vectors[bb_id]*scales[bb_id]
            )

        return new_position_matrix

    def _get_new_cell_vectors(self, unit_cell, step, scales):
        """
        Get the unit cell vectors after collapse step.

        """

        vector_1 = unit_cell.get_vector_1()
        vector_2 = unit_cell.get_vector_2()
        vector_3 = unit_cell.get_vector_3()

        max_scale = max(list(scales.values()))

        vector_1 = vector_1 - vector_1*step*max_scale
        vector_2 = vector_2 - vector_2*step*max_scale
        vector_3 = vector_3 - vector_3*step*max_scale

        return vector_1, vector_2, vector_3

    def _get_bb_vectors(self, mol, bb_atom_ids):
        """
        Get the building block to COM vectors.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        bb_atom_ids : :class:`dict` mapping :class:`int`: :class:`list`
            Dictionary mapping building block ids (keys) to a list of
            atom ids (values) in each distinct building block in the
            molecule.

        Returns
        -------
        bb_cent_vectors :
            :class:`dict` mapping :class:`int`: :class:`numpy.ndarray`
            Dictionary mapping building block ids (keys) to centroid
            vectors (values) of each distinct building block in the
            molecule.

        bb_cent_scales :
            :class:`dict` mapping :class:`int`: :class:`float`
            Dictionary mapping building block ids (keys) to relative
            magnitude of centroid vectors (values) of each distinct
            building block in the molecule.

        """

        cent = mol.get_centroid()

        # Get bb COM vector to molecule COM.
        bb_cent_vectors = {
            i: mol.get_centroid(atom_ids=bb_atom_ids[i])-cent
            for i in bb_atom_ids
        }

        # Scale the step size based on the different distances of
        # bbs from the COM. Impacts anisotropic topologies.
        if self._scale_steps:
            norms = {
                i: np.linalg.norm(bb_cent_vectors[i])
                for i in bb_cent_vectors
            }
            max_distance = max(list(norms.values()))
            bb_cent_scales = {
                i: norms[i]/max_distance
                for i in norms
            }
        else:
            bb_cent_scales = {
                i: 1
                for i in bb_cent_vectors
            }

        return bb_cent_vectors, bb_cent_scales

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`stk.ConstructedMolecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`stk.ConstructedMolecule`
            The optimized molecule.

        """

        # Handle output dir.
        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        os.mkdir(output_dir)

        bb_atom_ids = defaultdict(list)
        for i in mol.get_atom_infos():
            bb_atom_ids[i.get_building_block_id()].append(
                i.get_atom().get_id()
            )

        # Translate each building block along bb_COM_vectors by a
        # distance `step`. I.e. `step` is the proportion of the
        # bb_COM_vectors that the building block is moved.
        step_no = 0
        step = self._step_size
        while not self._has_short_contacts(mol):
            # Update each step the building block vectors and distance.
            bb_cent_vectors, bb_cent_scales = self._get_bb_vectors(
                mol=mol,
                bb_atom_ids=bb_atom_ids
            )

            new_pos = self._get_new_position_matrix(
                mol=mol,
                step=step,
                vectors=bb_cent_vectors,
                scales=bb_cent_scales
            )
            step_no += 1
            mol = mol.with_position_matrix(new_pos)
            mol.write(
                os.path.join(output_dir, f'collapsed_{step_no}.mol')
            )

        bb_cent_vectors, bb_cent_scales = self._get_bb_vectors(
            mol=mol,
            bb_atom_ids=bb_atom_ids
        )

        # Check that we have not gone too far.
        min_dist = min(
            dist for dist in self._get_inter_bb_distance(mol)
        )
        if min_dist < self._distance_cut / 2:
            # Revert to half the previous step if we have.
            step = -(self._step_size/2)
            new_pos = self._get_new_position_matrix(
                mol=mol,
                step=step,
                vectors=bb_cent_vectors,
                scales=bb_cent_scales
            )
            step_no += 1
            mol = mol.with_position_matrix(new_pos)
            mol.write(
                os.path.join(output_dir, 'collapsed_rev.mol')
            )

        out_file = os.path.join(output_dir, 'collapser.out')
        with open(out_file, 'w') as f:
            f.write(
                "Collapser algorithm.\n"
                "====================\n"
                f"Step size: {self._step_size}\n"
                f"Scale steps?: {self._scale_steps}\n"
                f"Distance cut: {self._distance_cut}\n"
                f"====================\n"
                f"Steps run: {step_no}\n"
                f"Minimum inter-bb distance: {min_dist}\n"
                f"====================\n"
            )
        return mol

    def p_optimize(self, mol, unit_cell):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`stk.ConstructedMolecule`
            The molecule to be optimized.

        unit_cell : :class:`.UnitCell`
            The cell to be optimized.

        Returns
        -------
        mol : :class:`stk.ConstructedMolecule`
            The optimized molecule.

        unit_cell : :class:`.UnitCell`
            The optimized cell.

        """

        # Handle output dir.
        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        os.mkdir(output_dir)

        bb_atom_ids = defaultdict(list)
        for i in mol.get_atom_infos():
            bb_atom_ids[i.get_building_block_id()].append(
                i.get_atom().get_id()
            )

        # Translate each building block along bb_COM_vectors by a
        # distance `step`. I.e. `step` is the proportion of the
        # bb_COM_vectors that the building block is moved.
        step_no = 0
        step = self._step_size
        while not self._has_short_contacts(mol):
            # Update each step the building block vectors and distance.
            bb_cent_vectors, bb_cent_scales = self._get_bb_vectors(
                mol=mol,
                bb_atom_ids=bb_atom_ids
            )

            new_pos = self._get_new_position_matrix(
                mol=mol,
                step=step,
                vectors=bb_cent_vectors,
                scales=bb_cent_scales,
            )
            step_no += 1
            mol = mol.with_position_matrix(new_pos)
            new_vector_1, new_vector_2, new_vector_3 = (
                self._get_new_cell_vectors(
                    unit_cell=unit_cell,
                    step=step,
                    scales=bb_cent_scales,
                )
            )
            unit_cell = unit_cell.with_cell_from_vectors(
                vector_1=new_vector_1,
                vector_2=new_vector_2,
                vector_3=new_vector_3,
            )
            PdbWriter().write(
                molecule=mol,
                path=os.path.join(
                    output_dir, f'collapsed_{step_no}.pdb'
                ),
                periodic_info=unit_cell,
            )

        bb_cent_vectors, bb_cent_scales = self._get_bb_vectors(
            mol=mol,
            bb_atom_ids=bb_atom_ids
        )

        # Check that we have not gone too far.
        min_dist = min(
            dist for dist in self._get_inter_bb_distance(mol)
        )
        if min_dist < self._distance_cut / 2:
            # Revert to half the previous step if we have.
            step = -(self._step_size/2)
            new_pos = self._get_new_position_matrix(
                mol=mol,
                step=step,
                vectors=bb_cent_vectors,
                scales=bb_cent_scales
            )
            step_no += 1
            mol = mol.with_position_matrix(new_pos)
            new_vector_1, new_vector_2, new_vector_3 = (
                self._get_new_cell_vectors(
                    unit_cell=unit_cell,
                    step=step,
                    scales=bb_cent_scales,
                )
            )
            unit_cell = unit_cell.with_cell_from_vectors(
                vector_1=new_vector_1,
                vector_2=new_vector_2,
                vector_3=new_vector_3,
            )
            PdbWriter().write(
                molecule=mol,
                path=os.path.join(
                    output_dir, 'collapsed_rev.pdb'
                ),
                periodic_info=unit_cell,
            )

        out_file = os.path.join(output_dir, 'collapser.out')
        with open(out_file, 'w') as f:
            f.write(
                "Collapser algorithm.\n"
                "====================\n"
                f"Step size: {self._step_size}\n"
                f"Scale steps?: {self._scale_steps}\n"
                f"Distance cut: {self._distance_cut}\n"
                f"====================\n"
                f"Steps run: {step_no}\n"
                f"Minimum inter-bb distance: {min_dist}\n"
                f"====================\n"
            )
        return mol, unit_cell


class CollapserMC(Collapser):
    """
    Collapse molecule to decrease enlarged bonds using MC algorithm.

    It is recommended to use the MCHammer version of this code with
    :mod:`MCHammer` [1]_, where a much cleaner version is written.
    The utilities `get_long_bond_ids` will help generate sub units.

    Smarter optimisation than Collapser using simple Monte Carlo
    algorithm to perform rigid translations of building blocks.

    References
    ----------
    .. [1] https://github.com/andrewtarzia/MCHammer

    """

    def __init__(
        self,
        output_dir,
        step_size,
        target_bond_length,
        num_steps,
        bond_epsilon=50,
        nonbond_epsilon=20,
        nonbond_sigma=1.2,
        nonbond_mu=3,
        beta=2,
        random_seed=None,
    ):
        """
        Initialize a :class:`Collapser` instance.

        Parameters
        ----------
        output_dir : :class:`str`
            The name of the directory into which files generated during
            the calculation are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        step_size : :class:`float`
            The relative size of the step to take during step.

        target_bond_length : :class:`float`
            Target equilibrium bond length for long bonds to minimize
            to.

        num_steps : :class:`int`
            Number of MC moves to perform.

        bond_epsilon : :class:`float`, optional
            Value of epsilon used in the bond potential in MC moves.
            Determines strength of the bond potential.
            Defaults to 50.

        nonbond_epsilon : :class:`float`, optional
            Value of epsilon used in the nonbond potential in MC moves.
            Determines strength of the nonbond potential.
            Defaults to 20.

        nonbond_sigma : :class:`float`, optional
            Value of sigma used in the nonbond potential in MC moves.
            Defaults to 1.2.

        nonbond_mu : :class:`float`, optional
            Value of mu used in the nonbond potential in MC moves.
            Determines the steepness of the nonbond potential.
            Defaults to 3.

        beta : :class:`float`, optional
            Value of beta used in the in MC moves. Beta takes the
            place of the inverse boltzmann temperature.
            Defaults to 2.

        random_seed : :class:`int`, optional
            Random seed to use for MC algorithm. Should only be set
            if exactly reproducible results are required, otherwise
            a system-based random seed should be used for proper
            sampling.

        """

        self._output_dir = output_dir
        self._step_size = step_size
        self._target_bond_length = target_bond_length
        self._num_steps = num_steps
        self._bond_epsilon = bond_epsilon
        self._nonbond_epsilon = nonbond_epsilon
        self._nonbond_sigma = nonbond_sigma
        self._nonbond_mu = nonbond_mu
        self._beta = beta
        if random_seed is None:
            random.seed()
        else:
            random.seed(random_seed)

    def _get_bb_atom_ids(self, mol):

        bb_atom_ids = defaultdict(list)
        for i in mol.get_atom_infos():
            bb_atom_ids[i.get_building_block_id()].append(
                i.get_atom().get_id()
            )

        return bb_atom_ids

    def _get_bond_length(self, mol, bond):

        position_matrix = mol.get_position_matrix()
        return get_atom_distance(
            position_matrix=position_matrix,
            atom1_id=bond.get_atom1().get_id(),
            atom2_id=bond.get_atom2().get_id()
        )

    def _get_bond_vector(self, mol, bond):

        position_matrix = mol.get_position_matrix()
        atom1_pos = position_matrix[bond.get_atom1().get_id()]
        atom2_pos = position_matrix[bond.get_atom2().get_id()]
        return atom2_pos - atom1_pos

    def _get_long_bond_infos(self, mol):
        """
        Returns dict of long bond infos.

        """

        long_bond_infos = {}
        for bond_infos in mol.get_bond_infos():
            if bond_infos.get_building_block() is None:
                ids = (
                    bond_infos.get_bond().get_atom1().get_id(),
                    bond_infos.get_bond().get_atom2().get_id(),
                )
                long_bond_infos[ids] = bond_infos

        return long_bond_infos

    def _get_bb_centroids(self, mol, bb_atom_ids):
        """
        Returns dict of building block centroids.

        """

        bb_centroids = {
            i: mol.get_centroid(atom_ids=bb_atom_ids[i])
            for i in bb_atom_ids
        }

        return bb_centroids

    def _get_cent_to_lb_vector(
        self,
        mol,
        bb_centroids,
        long_bond_infos
    ):
        """
        Returns dict of long bond atom to bb centroid vectors.

        """

        position_matrix = mol.get_position_matrix()
        centroid_to_lb_vectors = {}
        for bb in bb_centroids:
            cent = bb_centroids[bb]
            for b_atom_ids, bond_info in long_bond_infos.items():
                for atom_id in b_atom_ids:
                    atom_info, = mol.get_atom_infos(atom_ids=atom_id)
                    atom_pos = position_matrix[atom_id]
                    if atom_info.get_building_block_id() == bb:
                        centroid_to_lb_vectors[(bb, atom_id)] = (
                            atom_pos - cent,
                        )
                        break

        return centroid_to_lb_vectors

    def _bond_potential(self, distance):
        """
        Define an arbitrary parabolic bond potential.

        This potential has no relation to an empircal forcefield.
        """

        potential = (distance - self._target_bond_length) ** 2
        potential = self._bond_epsilon * potential

        return potential

    def _non_bond_potential(self, distance):
        """
        Define an arbitrary repulsive nonbonded potential.

        This potential has no relation to an empircal forcefield.
        """

        return (
            self._nonbond_epsilon * (
                (self._nonbond_sigma/distance) ** self._nonbond_mu
            )
        )

    def _compute_non_bonded_potential(self, mol):

        # Get all pairwise distances.
        pair_dists = pdist(mol.get_position_matrix())
        nonbonded_potential = np.sum(
            self._non_bond_potential(pair_dists)
        )

        return nonbonded_potential

    def _compute_potential(self, mol, long_bond_infos):

        system_potential = self._compute_non_bonded_potential(mol)
        for long_bond_ids in long_bond_infos:
            long_bond = long_bond_infos[long_bond_ids]

            system_potential += self._bond_potential(
                distance=self._get_bond_length(
                    mol=mol,
                    bond=long_bond.get_bond()
                )
            )

        return system_potential

    def _translate_atoms_along_vector(self, mol, atom_ids, vector):

        new_position_matrix = mol.get_position_matrix()
        for atom in mol.get_atom_infos(atom_ids=atom_ids):
            _id = atom.get_atom().get_id()
            pos = mol.get_position_matrix()[_id]
            new_position_matrix[_id] = pos - vector

        return mol.with_position_matrix(new_position_matrix)

    def _test_move(self, curr_pot, new_pot):

        if new_pot < curr_pot:
            return True
        else:
            exp_term = np.exp(-self._beta*(new_pot-curr_pot))
            rand_number = random.random()

            if exp_term > rand_number:
                return True
            else:
                return False

    def _output_top_lines(self):

        string = (
            '====================================================\n'
            '                Collapser optimisation              \n'
            '                ----------------------              \n'
            '                                                    \n'
            f' step size = {self._step_size} \n'
            f' target bond length = {self._target_bond_length} \n'
            f' num. steps = {self._num_steps} \n'
            f' bond epsilon = {self._bond_epsilon} \n'
            f' nonbond epsilon = {self._nonbond_epsilon} \n'
            f' nonbond sigma = {self._nonbond_sigma} \n'
            f' nonbond mu = {self._nonbond_mu} \n'
            f' beta = {self._beta} \n'
            '====================================================\n\n'
        )

        return string

    def _plot_progess(self, steps, maxds, spots, npots, output_dir):

        fig, ax = plt.subplots(figsize=(8, 5))
        ax.plot(steps, maxds, c='k', lw=2)
        # Set number of ticks for x-axis
        ax.tick_params(axis='both', which='major', labelsize=16)
        ax.set_xlabel('step', fontsize=16)
        ax.set_ylabel('max long bond length [angstrom]', fontsize=16)
        ax.axhline(y=self._target_bond_length, c='r')
        fig.tight_layout()
        fig.savefig(
            os.path.join(output_dir, 'maxd_vs_step.pdf'),
            dpi=360,
            bbox_inches='tight'
        )
        plt.close()
        # Plot energy vs timestep.
        fig, ax = plt.subplots(figsize=(8, 5))
        ax.plot(steps, spots, c='k', lw=2, label='total')
        ax.plot(steps, npots, c='r', lw=2, label='non bonded')
        # Set number of ticks for x-axis
        ax.tick_params(axis='both', which='major', labelsize=16)
        ax.set_xlabel('step', fontsize=16)
        ax.set_ylabel('potential', fontsize=16)
        ax.legend(fontsize=16)
        fig.tight_layout()
        fig.savefig(
            os.path.join(output_dir, 'pot_vs_step.pdf'),
            dpi=360,
            bbox_inches='tight'
        )
        plt.close()

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`stk.ConstructedMolecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`stk.ConstructedMolecule`
            The optimized molecule.

        """

        # Handle output dir.
        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        os.mkdir(output_dir)

        # Define long bonds to optimise.
        long_bond_infos = self._get_long_bond_infos(mol)

        # If no long bonds, then optimisation is done.
        if len(long_bond_infos) == 0:
            return mol

        # Define bb centroid - long bond atom vectors.
        # These are to be maintained during optimisation.
        # centroid_to_lb_vectors = self._get_cent_to_lb_vector(
        #     mol, bb_centroids, long_bond_infos
        # )
        print('###################################################')
        print('WARNING: centroid_to_lb_vectors not maintained yet.')
        print('###################################################')

        system_potential = self._compute_potential(
            mol=mol,
            long_bond_infos=long_bond_infos
        )

        with open(os.path.join(output_dir, 'coll.out'), 'w') as f:
            f.write(self._output_top_lines())
            mol.write(os.path.join(output_dir, 'coll_0.mol'))
            steps = [0]
            passed = []
            spots = [system_potential]
            npots = [self._compute_non_bonded_potential(mol=mol)]
            maxds = [max([
                self._get_bond_length(
                    mol, long_bond_infos[i].get_bond()
                )
                for i in long_bond_infos
            ])]
            f.write(
                'Step system_potential nonbond_potential max_dist '
                'opt_bbs updated?\n'
            )
            f.write(
                f'{steps[-1]} {spots[-1]} {npots[-1]} {maxds[-1]} '
                '-- --\n'
            )
            for step in range(1, self._num_steps):

                # Randomly select a long bond.
                lb_ids = random.choice(list(long_bond_infos.keys()))
                lb_info = long_bond_infos[lb_ids]

                lb_vector = self._get_bond_vector(
                    mol=mol,
                    bond=lb_info.get_bond()
                )

                bb_id_1, bb_id_2 = (
                    i.get_building_block_id()
                    for i in mol.get_atom_infos(atom_ids=lb_ids)
                )

                # Choose bb to move out of the two randomly.
                moving_bb = random.choice([bb_id_1, bb_id_2])
                moving_bb_atom_ids = tuple(
                    i.get_atom().get_id()
                    for i in mol.get_atom_infos()
                    if i.get_building_block_id() == moving_bb
                )

                # Randomly choose between translation along long bond
                # vector or along BB-COM vector.
                # Random number from -1 to 1
                rand = (random.random() - 0.5) * 2

                # Define translation along long bond vector where
                # direction is from force, magnitude is randomly
                # scaled.
                long_bond_translation = (
                    -lb_vector * self._step_size * rand
                )

                # Get bb COM vector to molecule COM.
                cent = mol.get_centroid()
                bb_cent_vector = (
                    mol.get_centroid(atom_ids=moving_bb_atom_ids)-cent
                )
                com_translation = (
                    bb_cent_vector * self._step_size * rand
                )

                translation_vector = random.choice([
                    long_bond_translation,
                    com_translation,
                ])

                # Translate building block.
                # Update atom position of building block.
                mol = self._translate_atoms_along_vector(
                    mol=mol,
                    atom_ids=moving_bb_atom_ids,
                    vector=translation_vector,
                )

                ###################################################
                # Here I want to add rotations
                # To maintain cent_vectors relative orientations.
                # cent_vector_1 = centroid_to_lb_vectors[
                #     (bb_id_1, lb_ids[0])
                # ]
                # cent_vector_2 = centroid_to_lb_vectors[
                #     (bb_id_2, lb_ids[1])
                # ]
                ###################################################

                new_system_potential = self._compute_potential(
                    mol=mol,
                    long_bond_infos=long_bond_infos
                )

                if self._test_move(
                    system_potential,
                    new_system_potential
                ):
                    updated = 'T'
                    system_potential = new_system_potential
                    passed.append(step)
                else:
                    updated = 'F'
                    # Reverse move.
                    mol = self._translate_atoms_along_vector(
                        mol=mol,
                        atom_ids=moving_bb_atom_ids,
                        vector=-translation_vector,
                    )

                mol.write(os.path.join(output_dir, f'coll_{step}.xyz'))
                steps.append(step)
                spots.append(system_potential)
                npots.append(
                    self._compute_non_bonded_potential(mol=mol)
                )
                maxds.append(max([
                    self._get_bond_length(
                        mol, long_bond_infos[i].get_bond()
                    )
                    for i in long_bond_infos
                ]))
                f.write(
                    f'{steps[-1]} {spots[-1]} '
                    f'{npots[-1]} {maxds[-1]} {lb_ids} {updated}\n'
                )
                step += 1

            f.write('\n============================================\n')
            f.write(
                'Optimisation done:\n'
                f'{len(passed)} steps passed: '
                f'{len(passed)/self._num_steps}'
            )

        self._plot_progess(steps, maxds, spots, npots, output_dir)

        return mol



"""
Molecule Splitter
=================

#. :class:`.MoleculeSplitter`

Class for splitting a molecule into many with new connectors.

"""

import logging
from rdkit.Chem import AllChem as rdkit
from itertools import combinations

import stk

logger = logging.getLogger(__name__)


class MoleculeSplitter:
    """
    Split an stk.molecule into many with new functional groups.

    Examples
    --------

    Given a molecule, this class allows you to break bonds based on
    `breaker_smarts` between the atoms in `bond_deleter_ids` and cap
    the broken bond by an atom defined in `replacer_smarts`.

    .. code-block:: python

        import stk
        import stko

        full_mol = stk.BuildingBlock('C1=CC=NC(=C1)C=NC2=CC=C(C=C2)Br')

        splitter = stko.MoleculeSplitter(
            breaker_smarts='[#6X3]~[#7X2]~[#6X3H1]~[#6X3!H1]',
            bond_deleter_ids=(0, 1),
            replacer_smarts='[Br]',
            functional_groups=(stk.BromoFactory(), ),
        )
        split_mols = splitter.split(full_mol)

    """

    def __init__(
        self,
        breaker_smarts,
        bond_deleter_ids,
        replacer_smarts,
        functional_groups,
    ):
        """
        Initialize a :class:`MoleculeSplitter`.

        Parameters
        ----------
        breaker_smarts : :class:`str`
            SMARTS string to find the substructure to break.

        bond_deleter_ids : :class:`tuple` of :class:`int`
            Index of atoms in `breaker_smarts` to breka bond between.

        replacer_smarts : :class:`str`
            SMARTS string of atom to replace dummy atoms with. This
            must be a single atom.

        functional_groups :
        :class:`iterable` of :class:`stk.FunctionalGroupFactory`
            Functional group factories to use to define new building
            block.

        """

        self._breaker_smarts = breaker_smarts
        self._bond_deleter_ids = bond_deleter_ids
        self._replacer_smarts = replacer_smarts
        self._functional_groups = functional_groups

    def split(self, molecule):
        """
        Split a molecule into multiple molecules.

        Parameters
        ----------
        molecule : :class:`stk.Molecule`
            Molecule to split.

        Returns
        -------
        :class:`iterable` of :class:`stk.BuildingBlock`
            The resulting list of building blocks, with new functional
            groups assigned.

        """

        new_building_blocks = []

        rdkit_mol = molecule.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_mol)
        bond_pair_ids_to_delete = []
        reactable_atom_ids = []
        for atom_ids in rdkit_mol.GetSubstructMatches(
            query=rdkit.MolFromSmarts(self._breaker_smarts)
        ):
            translated_bond_atom_ids = (
                atom_ids[self._bond_deleter_ids[0]],
                atom_ids[self._bond_deleter_ids[1]],
            )
            bond_pair_ids_to_delete.extend(
                tuple(sorted((i, j)))
                for i, j in combinations(translated_bond_atom_ids, 2)
            )
            reactable_atom_ids.extend(
                i for i in translated_bond_atom_ids
            )

        bond_ids_to_delete = []
        for bond in rdkit_mol.GetBonds():
            idxs = tuple(sorted((
                bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            )))
            if idxs in bond_pair_ids_to_delete:
                bond_ids_to_delete.append(bond.GetIdx())

        # Delete bonds and atoms.
        fragments = rdkit.GetMolFrags(rdkit.FragmentOnBonds(
            mol=rdkit_mol,
            bondIndices=tuple(bond_ids_to_delete),
            addDummies=True,
        ), asMols=True)

        # Add replacers bonded to * atoms.
        for i in rdkit.MolFromSmarts(self._replacer_smarts).GetAtoms():
            replacer_atom = rdkit.Atom(i.GetAtomicNum())
            replacer_atom.SetFormalCharge(0)
            break

        for frag in fragments:
            editable = rdkit.EditableMol(frag)
            for atom in frag.GetAtoms():
                # 0 if dummy atom.
                if atom.GetAtomicNum() == 0:
                    editable.ReplaceAtom(atom.GetIdx(), replacer_atom)

            frag_rdkit = editable.GetMol()
            rdkit.Kekulize(frag_rdkit)
            new_building_blocks.append(
                stk.BuildingBlock.init_from_rdkit_mol(
                    molecule=frag_rdkit,
                    functional_groups=self._functional_groups
                )
            )

        return new_building_blocks



import stko
from .utilities import (
    inequivalent_position_matrices,
    is_equivalent_molecule,
)


def test_MMFF_opt(unoptimized_mol):
    optimizer = stko.MMFF()
    opt_res = optimizer.optimize(unoptimized_mol)
    is_equivalent_molecule(opt_res, unoptimized_mol)
    inequivalent_position_matrices(opt_res, unoptimized_mol)


def test_UFF_opt(unoptimized_mol):
    optimizer = stko.UFF()
    opt_res = optimizer.optimize(unoptimized_mol)
    is_equivalent_molecule(opt_res, unoptimized_mol)
    inequivalent_position_matrices(opt_res, unoptimized_mol)


def test_ETKDG_opt(unoptimized_mol):
    optimizer = stko.ETKDG()
    opt_res = optimizer.optimize(unoptimized_mol)
    is_equivalent_molecule(opt_res, unoptimized_mol)
    inequivalent_position_matrices(opt_res, unoptimized_mol)

    optimizer = stko.ETKDG(random_seed=2584)
    opt_res = optimizer.optimize(unoptimized_mol)
    is_equivalent_molecule(opt_res, unoptimized_mol)
    inequivalent_position_matrices(opt_res, unoptimized_mol)



"""
Optimizers
==========

#. :class:`.Optimizer`
#. :class:`.OptimizerSequence`
#. :class:`.TryCatchOptimizer`


Optimizers are objects used to optimize molecules. Each optimizer is
initialized with some settings and can optimize a molecule
with :meth:`~.Optimizer.optimize`.

.. code-block:: python

    import stk
    import stko

    mol = stk.BuildingBlock('NCCCN', [stk.PrimaryAminoFactory()])
    mmff = stko.MMFF()
    mol = mmff.optimize(mol)

    # Optimizers also work with ConstructedMolecule objects.
    polymer = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(mol, ),
            repeating_unit='A',
            num_repeating_units=3,
        )
    )
    etkdg = stko.ETKDG()
    polymer = etkdg.optimize(polymer)

Sometimes it is desirable to chain multiple optimizations, one after
another. For example, before running an optimization, it may be
desirable to embed a molecule first, to generate an initial structure.
:class:`.OptimizerSequence` may be used for this.

.. code-block:: python

    # Create a new optimizer which chains the previously defined
    # mmff and etkdg optimizers.
    optimizer_sequence = stko.OptimizerSequence(etkdg, mmff)

    # Run each optimizer in sequence.
    polymer = optimizer_sequence.optimize(polymer)

.. _`adding optimizers`:

Making New Optimizers
---------------------

New optimizers can be made by simply making a class which inherits the
:class:`.Optimizer` class. In addition to this, the new class must
define a :meth:`~.Optimizer.optimize` method. The method must take 1
mandatory `mol` parameter. :meth:`~.Optimizer.optimize` will take the
`mol` and change its structure in whatever way it likes. Beyond this
there are no requirements. New optimizers can be added into the
:mod:`.optimizers` submodule or into a new submodule.

"""

import logging


logger = logging.getLogger(__name__)


class Optimizer:
    """
    A base class for optimizers.

    """

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        ...


class OptimizerSequence(Optimizer):
    """
    Applies optimizers in sequence.

    Examples
    --------
    Let's say we want to embed a molecule with ETKDG first and then
    minimize it with the MMFF force field.

    .. code-block:: python

        import stk
        import stko

        mol = stk.BuildingBlock('NCCCN', [stk.PrimaryAminoFactory()])
        optimizer = stko.OptimizerSequence(stko.ETKDG(), stko.MMFF())
        mol = optimizer.optimize(mol)

    """

    def __init__(self, *optimizers):
        """
        Initialize a :class:`OptimizerSequence` instance.

        Parameters
        ----------
        *optimizers : :class:`Optimizer`
            A number of optimizers, each of which gets applied to a
            molecule, based on the order given.

        """

        self._optimizers = optimizers

    def optimize(self, mol):

        for optimizer in self._optimizers:
            cls_name = optimizer.__class__.__name__
            logger.info(f'Using {cls_name} on "{mol}".')
            mol = optimizer.optimize(mol)

        return mol


class TryCatchOptimizer(Optimizer):
    """
    Try to optimize with a :class:`Optimizer`, use another on failure.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko

        # Create some molecules to optimize.
        mol1 = stk.BuildingBlock('NCCN')
        mol2 = stk.BuildingBlock('CCCCC')
        mol3 = stk.BuildingBlock('O=CCCN')

        # Create an optimizer which may fail.
        uff = stko.UFF()

        # Create a backup optimizer.
        mmff = stko.MMFF()

        # Make an optimizer which tries to run raiser and if that
        # raises an error, will run mmff on the molecule instead.
        try_catch = stko.TryCatchOptimizer(
            try_optimizer=uff,
            catch_optimizer=mmff,
        )

        # Optimize the molecules. In each case if the optimization with
        # UFF fails, MMFF is used to optimize the molecule instead.
        mol1 = try_catch.optimize(mol1)
        mol2 = try_catch.optimize(mol2)
        mol3 = try_catch.optimize(mol3)

    """

    def __init__(self, try_optimizer, catch_optimizer):
        """
        Initialize a :class:`TryCatchOptimizer` instance.

        Parameters
        ----------
        try_optimizer : :class:`Optimizer`
            The optimizer which is used initially to try and optimize a
            :class:`.Molecule`.

        catch_optimizer : :class:`Optimizer`
            If `try_optimizer` raises an error, this optimizer is
            run on the :class:`.Molecule` instead.

        """

        self._try_optimizer = try_optimizer
        self._catch_optimizer = catch_optimizer

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        """

        try:
            return self._try_optimizer.optimize(mol)
        except Exception:
            try_name = self._try_optimizer.__class__.__name__
            catch_name = self._catch_optimizer.__class__.__name__
            logger.error(
                f'{try_name} failed, trying {catch_name}.',
                exc_info=True
            )
            return self._catch_optimizer.optimize(mol)



"""
RDKit Optimizers
================

#. :class:`.MMFF`
#. :class:`.UFF`
#. :class:`.ETKDG`
#. :class:`.MetalOptimizer`

Wrappers for optimizers within the :mod:`rdkit` code.

.. code-block:: python

    import stk
    import stko

    mol = stk.BuildingBlock('NCCCN', [stk.PrimaryAminoFactory()])

    # Optimizers work on stk.Molecule.
    polymer = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(mol, ),
            repeating_unit='A',
            num_repeating_units=3,
        )
    )
    etkdg = stko.ETKDG()
    polymer = etkdg.optimize(polymer)

"""

import logging
import numpy as np
import rdkit.Chem.AllChem as rdkit
from itertools import combinations

from .optimizers import Optimizer
from ..utilities import (
    get_metal_atoms,
    get_metal_bonds,
    to_rdkit_mol_without_metals,
    vector_angle,
)

logger = logging.getLogger(__name__)


class MMFF(Optimizer):
    """
    Use the MMFF force field to optimize molecules.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko

        mol = stk.BuildingBlock('NCCNCCN')
        mmff = stko.MMFF()
        mol = mmff.optimize(mol)

    """

    def __init__(self, ignore_inter_interactions=True):

        self._ignore_inter_interactions = (
            ignore_inter_interactions
        )

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        rdkit_mol = mol.to_rdkit_mol()
        # Needs to be sanitized to get force field params.
        rdkit.SanitizeMol(rdkit_mol)
        rdkit.MMFFOptimizeMolecule(
            rdkit_mol,
            ignoreInterfragInteractions=self._ignore_inter_interactions
        )
        mol = mol.with_position_matrix(
            position_matrix=rdkit_mol.GetConformer().GetPositions()
        )

        return mol


class UFF(Optimizer):
    """
    Use the UFF force field to optimize molecules.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko

        mol = stk.BuildingBlock('NCCNCCN')
        uff = stko.UFF()
        mol = uff.optimize(mol)

    """

    def __init__(self, ignore_inter_interactions=True):

        self._ignore_inter_interactions = ignore_inter_interactions

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        rdkit_mol = mol.to_rdkit_mol()
        # Needs to be sanitized to get force field params.
        rdkit.SanitizeMol(rdkit_mol)
        rdkit.UFFOptimizeMolecule(
            rdkit_mol,
            ignoreInterfragInteractions=self._ignore_inter_interactions
        )
        mol = mol.with_position_matrix(
            position_matrix=rdkit_mol.GetConformer().GetPositions()
        )

        return mol


class ETKDG(Optimizer):
    """
    Uses the ETKDG [#]_ v2 algorithm to find an optimized structure.

    Examples
    --------
    .. code-block:: python

        import stk
        import stko

        mol = stk.BuildingBlock('NCCNCCN')
        etkdg = stko.ETKDG()
        mol = etkdg.optimize(mol)

    References
    ----------
    .. [#] http://pubs.acs.org/doi/pdf/10.1021/acs.jcim.5b00654

    """

    def __init__(self, random_seed=12):
        """
        Initialize a :class:`ETKDG` instance.

        Parameters
        ----------
        random_seed : :class:`int`, optional
            The random seed to use.

        """

        self._random_seed = random_seed

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """

        params = rdkit.ETKDGv2()
        params.clearConfs = True
        params.random_seed = self._random_seed

        rdkit_mol = mol.to_rdkit_mol()
        rdkit.EmbedMolecule(rdkit_mol, params)
        mol = mol.with_position_matrix(
            position_matrix=rdkit_mol.GetConformer().GetPositions()
        )

        return mol


class MetalOptimizer(Optimizer):
    """
    Applies forcefield optimizers that can handle metal centres.

    Notes
    -----
    By default, :meth:`optimize` will run a restricted optimization
    using constraints and the UFF. To implement this, metal atoms are
    replaced by noninteracting H atoms, and constraints are applied
    to maintain the metal centre geometry.
    Restrictions are applied to the ligand with respect to its input
    structure. So if that is poorly optimised, then the output will be
    also.

    Examples
    --------
    :class:`MetalOptimizer` allows for the restricted optimization of
    :class:`ConstructedMolecule` instances containing metals. Note that
    this optimizer algorithm is not very robust to large bonds and may
    fail.

    .. code-block:: python
        import stk
        import stko

        # Produce a Pd+2 atom with 4 functional groups.
        palladium_atom = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(4)
            ),
            position_matrix=[[0., 0., 0.]],
        )

        # Build a building block with two functional groups using
        # the SmartsFunctionalGroupFactory.
        bb1 = stk.BuildingBlock(
            smiles=(
                'C1=NC=CC(C2=CC=CC(C3=C'
                'C=NC=C3)=C2)=C1'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage1 = stk.ConstructedMolecule(
            stk.cage.M3L6(
                building_blocks=(palladium_atom, bb1),
                # Ensure that bonds between the GenericFunctionalGroups
                # of the ligand and the SingleAtom functional groups
                # of the metal are dative.
                reaction_factory=stk.DativeReactionFactory(
                    stk.GenericReactionFactory(
                        bond_orders={
                            frozenset({
                                stk.GenericFunctionalGroup,
                                stk.SingleAtom
                            }): 9
                        }
                    )
                ),
                optimizer=stk.MCHammer(),
            )
        )

        # Define an optimizer.
        optimizer = stko.MetalOptimizer()

        # Optimize.
        cage1 = optimizer.optimize(mol=cage1)

    """

    def __init__(
        self,
        metal_binder_distance=1.6,
        metal_binder_forceconstant=1.0e2,
        max_iterations=500,
    ):
        """
        Initialize a :class:`MetalOptimizer` instance.

        Parameters
        ----------
        metal_binder_distance : :class:`float`
            Distance in Angstrom.

        metal_binder_forceconstant : :class:`float`
            Force constant to use for restricted metal-ligand bonds.

        max_iterations : :class:`int`, optional
            Number of iteractions to run.

        """
        self._metal_binder_distance = metal_binder_distance
        self._metal_binder_forceconstant = metal_binder_forceconstant
        self._max_iterations = max_iterations

    def _apply_metal_centre_constraints(
        self,
        mol,
        ff,
        metal_bonds,
    ):
        """
        Applies UFF metal centre constraints.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        ff : :class:`rdkit.ForceField`
            Forcefield to apply constraints to. Generally use UFF.

        metal_bonds : :class:`.list` of :class:`stk.Bond`
            List of bonds including metal atoms.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Add constraints to UFF to hold metal geometry in place.
        for bond in metal_bonds:
            idx1 = bond.get_atom1().get_id()
            idx2 = bond.get_atom2().get_id()
            # Add distance constraints in place of metal bonds.
            # Target distance set to a given metal_binder_distance.
            ff.UFFAddDistanceConstraint(
                idx1=idx1,
                idx2=idx2,
                relative=False,
                minLen=self._metal_binder_distance,
                maxLen=self._metal_binder_distance,
                forceConstant=self._metal_binder_forceconstant
            )

        # Also implement angular constraints to all atoms in the
        # metal complex.
        for bonds in combinations(metal_bonds, r=2):
            bond1, bond2 = bonds
            bond1_atoms = [bond1.get_atom1(), bond1.get_atom2()]
            bond2_atoms = [bond2.get_atom1(), bond2.get_atom2()]
            pres_atoms = list(set(bond1_atoms + bond2_atoms))
            # If there are more than 3 atoms, implies two
            # independant bonds.
            if len(pres_atoms) > 3:
                continue
            for atom in pres_atoms:
                if atom in bond1_atoms and atom in bond2_atoms:
                    idx2 = atom.get_id()
                elif atom in bond1_atoms:
                    idx1 = atom.get_id()
                elif atom in bond2_atoms:
                    idx3 = atom.get_id()
            pos1 = [
                i for i in mol.get_atomic_positions(atom_ids=[idx1])
            ][0]
            pos2 = [
                i for i in mol.get_atomic_positions(atom_ids=[idx2])
            ][0]
            pos3 = [
                i for i in mol.get_atomic_positions(atom_ids=[idx3])
            ][0]
            v1 = pos1 - pos2
            v2 = pos3 - pos2
            angle = vector_angle(v1, v2)
            ff.UFFAddAngleConstraint(
                idx1=idx1,
                idx2=idx2,
                idx3=idx3,
                relative=False,
                minAngleDeg=np.degrees(angle),
                maxAngleDeg=np.degrees(angle),
                forceConstant=1.0e5
            )

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        """
        # Find all metal atoms and atoms they are bonded to.
        metal_atoms = get_metal_atoms(mol)
        metal_bonds, ids_to_metals = get_metal_bonds(
            mol=mol,
            metal_atoms=metal_atoms
        )

        # Perform a forcefield optimisation that
        # only optimises non metal atoms that are not bonded to the
        # metal.

        # Write rdkit molecule with metal atoms and bonds deleted.
        edit_mol = to_rdkit_mol_without_metals(
            mol=mol,
            metal_atoms=metal_atoms,
            metal_bonds=metal_bonds
        )

        # Non-bonded interactions need to be explicitly turned on (if
        # desired) because at the point of initialisation, the
        # molecules are technically separate (not bonded) and are
        # treated as fragments.
        rdkit.SanitizeMol(edit_mol)
        ff = rdkit.UFFGetMoleculeForceField(
            edit_mol,
            ignoreInterfragInteractions=False,
        )

        # Constrain the metal centre.
        self._apply_metal_centre_constraints(
            mol=mol,
            ff=ff,
            metal_bonds=metal_bonds,
        )

        # Optimisation with UFF in RDKit. This method uses constraints
        # on the metal centre to attempt to enforce the metal geometry
        # described by the metal topology.
        ff.Minimize(maxIts=self._max_iterations)

        # Update stk molecule from optimized molecule. This should
        # only modify atom positions, which means metal atoms will be
        # reinstated.
        new_position_matrix = edit_mol.GetConformer().GetPositions()
        mol = mol.with_position_matrix(new_position_matrix)

        return mol



"""
Extractor
=========

#. :class:`.Extractor`

Base results class for extracting molecule properties.

"""

from abc import ABC


class Extractor(ABC):
    """
    An abstract base class for extractors.

    """

    def __init__(self, output_file):
        """
        Initializes :class:`Extractor`

        Parameters
        ----------
        output_file : :class:`str`
            Output file to extract properties from.

        """

        self.output_file = output_file
        # Explictly set encoding to UTF-8 because default encoding on
        # Windows will fail to read the file otherwise.
        with open(self.output_file, 'r', encoding='UTF-8') as f:
            self.output_lines = f.readlines()

        self._extract_values()

    def _extract_values(self):
        """
        Extract values from `path`.

        Parameters
        ----------
        path : :class:`str`
            Path to file to extract values from.

        """

        raise NotImplementedError()

    def _properties_dict(self):
        """
        Dictionary of defined properties and extraction strings.

        """

        raise NotImplementedError()

    def _check_line(self, line, option):
        """
        Checks a line for a string based on option.

        Parameters
        ----------
        line : :class:`str`
            Line of output file to check.

        option : :class:`str`
            Define which property and string being checked for.
            They are defined in :meth:`_properties_dict`.

        Returns
        -------
        :class:`bool`
            Returns ``True`` if the desired string is present.

        """

        options = self._properties_dict()

        if options[option] in line:
            return True



"""
Energy Results
==============

#. :class:`.EnergyResults`

Results class for :class:`.Calculator` that outputs energy.

"""

from .results import Results


class EnergyResults(Results):
    """
    Results class containing molecule energy.

    """

    def __init__(self, generator, unit_string):

        self._value = next(generator)
        self._unit_string = unit_string

    def get_energy(self):
        return self._value

    def get_unit_string(self):
        return self._unit_string



"""
Torsion
=======

Class defining torsion angles.

"""

import logging

logger = logging.getLogger(__name__)


class Torsion:
    """
    Represents a torsion angle in a molecule.

    """

    def __init__(self, atom1, atom2, atom3, atom4):
        """
        Defines a torsion.

        Parameters
        ----------
        atom1 : :class:`stk.Atom`
            First atom in torsion.

        atom2 : :class:`stk.Atom`
            Second atom in torsion.

        atom3 : :class:`stk.Atom`
            Third atom in torsion.

        atom4 : :class:`stk.Atom`
            Fourth atom in torsion.

        """

        self._atom1 = atom1
        self._atom2 = atom2
        self._atom3 = atom3
        self._atom4 = atom4

    def get_atoms(self):
        return tuple((
            self._atom1,
            self._atom2,
            self._atom3,
            self._atom4,
        ))

    def get_atom_ids(self):
        return tuple(atom.get_id() for atom in self.get_atoms())

    def __iter__(self):
        return iter(self.get_atoms())

    def __str__(self):
        return (
            f'{self.__class__.__name__}({self._atom1}, '
            f'{self._atom2}, {self._atom3}, {self._atom4})'
        )

    def __repr__(self):
        return str(self)



"""
Results
=======

#. :class:`.Results`

Base results class for extracting molecule properties.

"""

from abc import ABC


class Results(ABC):
    """
    An abstract base class for results.

    """

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'<{self.__class__.__name__} at {id(self)}>'



"""
Z Matrix
========

#. :class:`.ZMatrix`

Class for converting a molecule to a Z-matrix.

"""

import logging

from ...utilities.utilities import (
    get_atom_distance,
    calculate_dihedral,
    calculate_angle,
)

logger = logging.getLogger(__name__)


class ConversionError(Exception):
    ...


class ZMatrix:
    """
    Converter for :class:`stk.Molecule` to Z-Matrix.

    Examples
    --------

    The Z-matrix is returned as a string.

    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('NCCNCCN')

        print(stko.ZMatrix().get_zmatrix(bb1))

    """

    def get_zmatrix(self, molecule):
        """
        Get Z-matrix of a molecule.

        Parameters
        ----------
        molecule : :class:`stk.Molecule`
            Molecule to convert.

        Returns
        -------
        zmatrix : :class:`str`
            The Z-matrix of the molecule.
            Distances in Angstrom, angles and torsions in degrees.

        """

        zmatrix = []
        mol_atoms = list(molecule.get_atoms())
        position_matrix = molecule.get_position_matrix()
        coords = 0
        for i, atom in enumerate(mol_atoms):
            if i == 0:
                zmatrix.append(f'{atom.__class__.__name__}')
            elif i == 1:
                distance = get_atom_distance(
                    position_matrix=position_matrix,
                    atom1_id=atom.get_id(),
                    atom2_id=mol_atoms[i-1].get_id(),
                )

                distance = round(distance, 2)
                zmatrix.append(
                    f'{atom.__class__.__name__} {i} {distance}'
                )
                coords += 1
            elif i == 2:
                distance = get_atom_distance(
                    position_matrix=position_matrix,
                    atom1_id=atom.get_id(),
                    atom2_id=mol_atoms[i-1].get_id(),
                )
                angle = calculate_angle(
                    pt1=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-3].get_id()
                        )
                    )[0],
                    pt2=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-2].get_id()
                        )
                    )[0],
                    pt3=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-1].get_id()
                        )
                    )[0],
                )

                distance = round(distance, 2)
                angle = round(angle, 2)
                zmatrix.append(
                    f'{atom.__class__.__name__} {i} {distance}'
                    f'{i-1} {angle}'
                )
                coords += 2

            else:
                distance = get_atom_distance(
                    position_matrix=position_matrix,
                    atom1_id=atom.get_id(),
                    atom2_id=mol_atoms[i-1].get_id(),
                )
                angle = calculate_angle(
                    pt1=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-3].get_id()
                        )
                    )[0],
                    pt2=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-2].get_id()
                        )
                    )[0],
                    pt3=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-1].get_id()
                        )
                    )[0],
                )
                torsion = calculate_dihedral(
                    pt1=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-3].get_id()
                        )
                    )[0],
                    pt2=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-2].get_id()
                        )
                    )[0],
                    pt3=tuple(
                        molecule.get_atomic_positions(
                            mol_atoms[i-1].get_id()
                        )
                    )[0],
                    pt4=tuple(
                        molecule.get_atomic_positions(atom.get_id())
                    )[0],
                )

                distance = round(distance, 2)
                angle = round(angle, 2)
                torsion = round(torsion, 2)
                zmatrix.append(
                    f'{atom.__class__.__name__} {i} {distance}'
                    f'{i-1} {angle} {i-2} {torsion}'
                )
                coords += 3

        if 3*len(zmatrix)-6 != coords:
            raise ConversionError(
                f'There are {coords}, not {3*len(zmatrix)-6} as '
                'expected. Therefore, the conversion has failed.'
            )

        return '\n'.join(zmatrix)



"""
Torsion Results
===============

#. :class:`.TorsionResults`
#. :class:`.ConstructedMoleculeTorsionResults`

Results classes for extracting molecular torsions.

"""

from collections import defaultdict
from .results import Results
from ...molecular.torsion import TorsionInfo, Torsion
from ...utilities import calculate_dihedral


class TorsionResults(Results):
    """
    Results class containing molecule torsions.

    """

    def __init__(self, generator, mol):
        self._torsions = next(generator)
        self._mol = mol

    def get_torsions(self):
        return self._torsions

    def get_molecule(self):
        return self._mol

    def get_torsion_angles(self):
        for torsion in self._torsions:
            print('a', torsion)
            yield (
                torsion, calculate_dihedral(
                    pt1=tuple(
                        self._mol.get_atomic_positions(
                            torsion.get_atom_ids()[0]
                        )
                    )[0],
                    pt2=tuple(
                        self._mol.get_atomic_positions(
                            torsion.get_atom_ids()[1]
                        )
                    )[0],
                    pt3=tuple(
                        self._mol.get_atomic_positions(
                            torsion.get_atom_ids()[2]
                        )
                    )[0],
                    pt4=tuple(
                        self._mol.get_atomic_positions(
                            torsion.get_atom_ids()[3]
                        )
                    )[0],
                )
            )


class ConstructedMoleculeTorsionResults(TorsionResults):
    """
    Results class containing molecule torsions.

    """

    def __init__(self, generator, mol):
        self._torsions = next(generator)
        self._mol = mol

    def get_torsion_infos_by_building_block(self):
        """
        Returns dictionary of torsions by building block.

        """

        torsion_infos_by_building_block = defaultdict(list)
        for torsion_info in self.get_torsion_infos():
            if torsion_info.get_building_block_id() is not None:
                torsion_infos_by_building_block[
                    torsion_info.get_building_block_id()
                ].append(torsion_info)
        return torsion_infos_by_building_block

    def get_torsion_infos(self):
        for torsion in self._torsions:
            atom_infos = list(
                self._mol.get_atom_infos(
                    atom_ids=(i for i in torsion.get_atom_ids())
                )
            )
            # Get atom info and check they are all the same.
            building_block_ids = set((
                i.get_building_block_id() for i in atom_infos
            ))
            if len(building_block_ids) > 1:
                same_building_block = False
            else:
                same_building_block = True

            if same_building_block:
                building_block_id = next(iter(building_block_ids))

                building_block = tuple((
                   i.get_building_block() for i in atom_infos
                ))[0]
                bb_atoms = tuple(
                    i.get_building_block_atom()
                    for i in atom_infos
                )
                building_block_torsion = Torsion(*bb_atoms)
                yield TorsionInfo(
                    torsion=torsion,
                    building_block=building_block,
                    building_block_id=building_block_id,
                    building_block_torsion=building_block_torsion,
                )
            else:
                yield TorsionInfo(
                    torsion=torsion,
                    building_block=None,
                    building_block_id=None,
                    building_block_torsion=None,
                )



from .results import *  # noqa
from .energy_results import *  # noqa
from .xtb_results import *  # noqa
from .torsion_results import *  # noqa
from .shape_results import *  # noqa
from .orca_results import *  # noqa



"""
Orca Calculators
================

#. :class:`.OrcaEnergy`

Wrappers for calculators within the :mod:`orca` code.

"""

import logging
import os
import shutil
import glob
import uuid
import subprocess as sp

from .calculators import Calculator
from .results import OrcaResults

logger = logging.getLogger(__name__)


class OrcaOptimizerError(Exception):
    ...


class OrcaEnergy(Calculator):
    """
    Uses Orca [1]_ to calculate energy and other properties.

    By default, :meth:`get_results` will extract other properties of
    the :class:`.Molecule` passed to :meth:`calculate`, which
    will be saved in the attributes of :class:`.OrcaResults`.

    All intermediate and output files from Orca are deleted at the end
    of the job (i.e. the `.gbw` file will be deleted) because they can
    quickly build up to large sizes. The `discard_output` option allows
    you to keep output files if desired.Additionally, the
    `write_input_only` option is available for jobs where you would
    like more customization or to run outside of the Python
    environment.

    Notes
    -----
    When running :meth:`calculate`, this calculator changes the
    present working directory with :func:`os.chdir`. The original
    working directory will be restored even if an error is raised, so
    unless multi-threading is being used this implementation detail
    should not matter.

    If multi-threading is being used an error could occur if two
    different threads need to know about the current working directory
    as :class:`.OrcaEnergy` can change it from under them.

    Note that this does not have any impact on multi-processing,
    which should always be safe.

    Examples
    --------

    .. code-block:: python

        import stk
        import stko

        bb1 = stk.BuildingBlock('NCCNCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit="AB",
                orientations=[0, 0],
                num_repeating_units=1
            )
        )

        # Optimize the constructed molecule so that it has a
        # reasonable structure.
        opt = stko.UFF()
        polymer = opt.optimize(polymer)

        # Calculate energy using Orca.
        orca = stko.OrcaEnergy(
            orca_path='/opt/orca/orca',
            topline='! SP B97-3c',
        )

        orca_results = orca.get_results(polymer)

        # Extract properties from the energy calculator for a given
        # molecule.
        total_energy = orca_results.get_total_energy()

    If you want the input file written (instead of the job run), you
    can use the `write_input_only` argument to save the input file
    in the `output_dir` as `orca_input.inp` with the input xyz file as
    `input_structure.xyz`.

    .. code-block:: python

        # Optimize the constructed molecule so that it has a
        # reasonable structure.
        optimizer = stko.ETKDG()
        polymer = optimizer.optimize(polymer)

        # Calculate energy using Orca.
        orca = stko.OrcaEnergy(
            orca_path='/opt/orca/orca',
            topline='! SP B97-3c',
            write_input_only=True,
        )

        orca.get_results(polymer)

    References
    ----------
    .. [1] https://orcaforum.kofo.mpg.de/app.php/portal

    """
    def __init__(
        self,
        orca_path,
        topline,
        basename=None,
        output_dir=None,
        num_cores=1,
        charge=0,
        multiplicity=1,
        write_input_only=False,
        discard_output=True,
    ):
        """
        Initializes a :class:`OrcaEnergy` instance.

        Parameters
        ----------
        orca_path : :class:`str`
            The path to the Orca executable.

        topline : :class:`str`
            Top line designating the type of calculation. Should start
            with `! `.

        basename : :class:`str`, optional
            Base name of Orca output files.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the calculation are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        num_cores : :class:`int`, optional
            The number of cores Orca should use.

        charge : :class:`int`, optional
            Formal molecular charge.

        multiplicity : :class:`int`, optional
            Multiplicity of system (2S+1), where S is the spin.

        write_input_only : :class:`bool`, optional
            `True` if you only want the input file written and to not
            have the Orca job run.

        discard_output : :class:`bool`, optional
            `True` if you want to delete auxillary Orca output files
            such as the `.gbw` file.

        """

        self._orca_path = orca_path
        if basename is None:
            self._basename = f'_{str(uuid.uuid4().int)}'
        else:
            self._basename = basename
        self._output_dir = output_dir
        self._topline = topline
        self._num_cores = str(num_cores)
        self._charge = str(charge)
        self._multiplicity = multiplicity
        self._write_input_only = write_input_only
        self._discard_output = discard_output

    def _write_input_file(self, path, xyz_file):
        # Write top line and base name.
        string = f'{self._topline}\n\n%base "{self._basename}"\n'

        # Add multiprocessing section.
        string += (
            f'%pal\n   nprocs {self._num_cores}\nend\n\n'
            f'%scf\n   MaxIter 2000\nend\n\n'
        )
        # Add geometry section.
        string += (
            f'* xyzfile {self._charge} {self._multiplicity} '
            f'{xyz_file}\n'
        )

        with open(path, 'w') as f:
            f.write(string)

    def _check_outcome(self, out_file):

        if not os.path.exists(out_file):
            raise OrcaOptimizerError(
                f'{out_file} does not exist, suggesting the job did '
                'not run.'
            )

        with open(out_file, 'r') as f:
            lines = f.readlines()
            if '****ORCA TERMINATED NORMALLY****' not in lines[-2]:
                raise OrcaOptimizerError(
                    'Orca job did not terminate normally.'
                )

        tmp_files = glob.glob(f'{self._basename}*tmp')
        if len(tmp_files) > 0:
            raise OrcaOptimizerError(
                'tmp files exist, suggesting the job did not complete '
                'or did not converge.'
            )

    def _clean_up(self):
        for to_del in glob.glob(f'{self._basename}*'):
            os.remove(to_del)

    def _run_orca(
        self,
        xyz_file,
        input_file,
        out_file,
        init_dir,
        output_dir,
    ):
        """
        Runs Orca.

        Parameters
        ----------
        xyz_file : :class:`str`
            The name of the input structure ``.xyz`` file.

        input_file : :class:`str`
            The name of input file to be written.

        out_file : :class:`str`
            The name of output file with Orca results.

        init_dir : :class:`str`
            The name of the current working directory.

        output_dir : :class:`str`
            The name of the directory into which files generated during
            the calculation are written.

        Returns
        -------
        None : :class:`NoneType`

        """

        cmd = f'{self._orca_path} {input_file}'

        try:
            os.chdir(output_dir)
            self._write_input_file(input_file, xyz_file)
            if not self._write_input_only:
                with open(out_file, 'w') as f:
                    # Note that sp.call will hold the program until
                    # completion of the calculation.
                    sp.call(
                        cmd,
                        stdin=sp.PIPE,
                        stdout=f,
                        stderr=sp.PIPE,
                        # Shell is required to run complex arguments.
                        shell=True
                    )
                self._check_outcome(out_file)
                if self._discard_output:
                    self._clean_up()
        finally:
            os.chdir(init_dir)

    def calculate(self, mol):
        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)
        os.mkdir(output_dir)

        init_dir = os.getcwd()
        xyz_file = os.path.join(output_dir, 'input_structure.xyz')
        input_file = os.path.join(output_dir, 'orca_input.inp')
        out_file = os.path.join(output_dir, 'orca_energy.output')
        mol.write(xyz_file)
        yield self._run_orca(
            xyz_file=xyz_file,
            input_file=input_file,
            out_file=out_file,
            init_dir=init_dir,
            output_dir=output_dir,
        )

    def get_results(self, mol):
        """
        Calculate the Orca properties of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`.OrcaResults`
            The properties, with units, from Orca calculations.

        """

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        out_file = os.path.join(output_dir, 'orca_energy.output')

        if self._write_input_only:
            next(self.calculate(mol))
        else:
            return OrcaResults(
                generator=self.calculate(mol),
                output_file=out_file,
            )

    def get_energy(self, mol):
        """
        Calculate the energy of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The :class:`.Molecule` whose energy is to be calculated.

        Returns
        -------
        :class:`float`
            The energy.

        """

        return self.get_results(mol).get_total_energy()[0]



import pytest

import stk
from stko import GulpUFFOptimizer, ExpectedMetal


pd_metal = stk.BuildingBlock(
    smiles='[Pd+2]',
    functional_groups=(
        stk.SingleAtom(stk.Pd(0, charge=2))
        for i in range(4)
    ),
    position_matrix=[[0, 0, 0]],
)

# Define a bidentate ligand with two functional groups.
bidentate_ligand = stk.BuildingBlock(
    smiles='NCCN',
    functional_groups=[
        stk.SmartsFunctionalGroupFactory(
            smarts='[#7]~[#6]',
            bonders=(0, ),
            deleters=(),
        ),
    ]
)

# Construct a cis-protected square planar metal complex.
complex = stk.ConstructedMolecule(
    stk.metal_complex.BidentateSquarePlanar(
        metals=pd_metal,
        ligands=bidentate_ligand,
    )
)


class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        The molecule to be tested.

    atom_types : :class:`dict`
        The smiles for the molecule.

    has_metal : :class:`bool`
        ``True`` if a metal atom is in molecule.

    metal_ff : :class:`dict` or :class:`NoneType`
        The position matrix of the molecule.

    """

    def __init__(self, molecule, atom_types, has_metal, metal_ff):
        self.molecule = molecule
        self.atom_types = atom_types
        self.has_metal = has_metal
        self.metal_ff = metal_ff


@pytest.fixture(
    params=(
        CaseData(
            molecule=stk.BuildingBlock('CCC'),
            atom_types={
                0: 'C_3', 1: 'C_3', 2: 'C_3',
                3: 'H_', 4: 'H_', 5: 'H_',
                6: 'H_', 7: 'H_',
                8: 'H_', 9: 'H_', 10: 'H_',
            },
            has_metal=False,
            metal_ff=None,
        ),
        # CaseData(
        #     molecule=stk.BuildingBlock.init_from_molecule(complex),
        #     atom_types={},
        #     has_metal=True,
        #     metal_ff=None,
        # ),
        CaseData(
            molecule=stk.BuildingBlock.init_from_molecule(complex),
            atom_types={
                0: 'Pd4+2',
                1: 'N_3', 2: 'C_3', 3: 'C_3', 4: 'N_3',
                5: 'H_', 6: 'H_', 7: 'H_', 8: 'H_', 9: 'H_',
                10: 'H_', 11: 'H_', 12: 'H_',
                13: 'N_3', 14: 'C_3', 15: 'C_3', 16: 'N_3',
                17: 'H_', 18: 'H_', 19: 'H_', 20: 'H_', 21: 'H_',
                22: 'H_', 23: 'H_', 24: 'H_',
            },
            has_metal=True,
            metal_ff={46: 'Pd4+2'},
        ),
        CaseData(
            molecule=stk.BuildingBlock('CCC'),
            atom_types={
                0: 'C_3', 1: 'C_3', 2: 'C_3',
                3: 'H_', 4: 'H_', 5: 'H_',
                6: 'H_', 7: 'H_',
                8: 'H_', 9: 'H_', 10: 'H_',
            },
            has_metal=False,
            metal_ff={26: 'Fe4+2'},
        ),
        CaseData(
            molecule=stk.BuildingBlock('c1ccccc1'),
            atom_types={
                0: 'C_R', 1: 'C_R', 2: 'C_R',
                3: 'C_R', 4: 'C_R', 5: 'C_R',
                6: 'H_', 7: 'H_', 8: 'H_', 9: 'H_', 10: 'H_', 11: 'H_',
            },
            has_metal=False,
            metal_ff=None,
        ),
        # CaseData(
        #     molecule=stk.BuildingBlock(
        # 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'
        # ),
        #     atom_types={
        #         0: 'C_3', 1: 'N_R', 2: 'C_R', 3: 'N_R', 4: 'C_R',
        #         5: 'C_R', 6: 'C_R', 7: 'O_2', 8: 'N_R', 9: 'C_R',
        #         10: 'O_2', 11: 'N_R', 12: 'C_3', 13: 'C_3',
        #         14: 'H_', 15: 'H_', 16: 'H_', 17: 'H_', 18: 'H_',
        #         19: 'H_', 20: 'H_', 21: 'H_', 22: 'H_', 23: 'H_',
        #     },
        #     has_metal=False,
        #     metal_ff=None,
        # ),
        CaseData(
            molecule=stk.BuildingBlock(
                'C1=CC(=CC(=C1)C#CC2=CN=CC=C2)C#CC3=CN=CC=C3'
            ),
            atom_types={
                0: 'C_R', 1: 'C_R', 2: 'C_R', 3: 'C_R', 4: 'C_R',
                5: 'C_R', 6: 'C_1', 7: 'C_1', 8: 'C_R', 9: 'C_R',
                10: 'N_R', 11: 'C_R', 12: 'C_R', 13: 'C_R', 14: 'C_1',
                15: 'C_1', 16: 'C_R', 17: 'C_R', 18: 'N_R', 19: 'C_R',
                20: 'C_R', 21: 'C_R', 22: 'H_', 23: 'H_', 24: 'H_',
                25: 'H_', 26: 'H_', 27: 'H_', 28: 'H_', 29: 'H_',
                30: 'H_', 31: 'H_', 32: 'H_', 33: 'H_',
            },
            has_metal=False,
            metal_ff=None,
        ),
        # CaseData(
        #     molecule=stk.BuildingBlock(
        #         'O=C1c2ccc(cc2Nc3cc(ccc13)C#Cc4cccnc4)C#Cc5cccnc5'
        #     ),
        #     atom_types={
        #         0: 'O_2', 1: 'C_R', 2: 'C_R', 3: 'C_R', 4: 'C_R',
        #         5: 'C_R', 6: 'C_R', 7: 'C_R', 8: 'N_R', 9: 'C_R',
        #         10: 'C_R', 11: 'C_R', 12: 'C_R', 13: 'C_R',
        # 14: 'C_R',
        #         15: 'C_1', 16: 'C_1', 17: 'C_R', 18: 'C_R',
        # 19: 'C_R',
        #         20: 'C_R', 21: 'N_R', 22: 'C_R', 23: 'C_1',
        # 24: 'C_1',
        #         25: 'C_R', 26: 'C_R', 27: 'C_R', 28: 'C_R',
        # 29: 'N_R',
        #         30: 'C_R', 31: 'H_', 32: 'H_', 33: 'H_', 34: 'H_',
        #         35: 'H_', 36: 'H_', 37: 'H_', 38: 'H_', 39: 'H_',
        #         40: 'H_', 41: 'H_', 42: 'H_', 43: 'H_', 44: 'H_',
        #         45: 'H_',
        #     },
        #     has_metal=False,
        #     metal_ff=None,
        # ),
        CaseData(
            molecule=stk.BuildingBlock('C1CCC(C(C1)N)N'),
            atom_types={
                0: 'C_3', 1: 'C_3', 2: 'C_3',
                3: 'C_3', 4: 'C_3', 5: 'C_3',
                6: 'N_3', 7: 'N_3',
                8: 'H_', 9: 'H_', 10: 'H_', 11: 'H_',
                12: 'H_', 13: 'H_', 14: 'H_', 15: 'H_',
                16: 'H_', 17: 'H_', 18: 'H_', 19: 'H_',
                20: 'H_', 21: 'H_',
            },
            has_metal=False,
            metal_ff=None,
        ),
        # CaseData(
        #     molecule=stk.BuildingBlock('C1=C(C=C(C=C1C=O)C=O)C=O'),
        #     atom_types={
        #         0: 'C_R', 1: 'C_R', 2: 'C_R',
        #         3: 'C_R', 4: 'C_R', 5: 'C_R',
        #         6: 'C_2', 7: 'O_2', 8: 'C_2',
        #         9: 'O_2', 10: 'C_2', 11: 'O_2',
        #         12: 'H_', 13: 'H_', 14: 'H_',
        #         15: 'H_', 16: 'H_', 17: 'H_',
        #     },
        #     has_metal=False,
        #     metal_ff=None,
        # ),
        CaseData(
            molecule=stk.BuildingBlock('CC=O'),
            atom_types={
                0: 'C_3', 1: 'C_2', 2: 'O_2',
                3: 'H_', 4: 'H_', 5: 'H_', 6: 'H_',
            },
            has_metal=False,
            metal_ff=None,
        ),
    ),
)
def test_molecule(request):
    return request.param


def test_assign_FF(test_molecule):
    gulp_opt = GulpUFFOptimizer(
        gulp_path='not_required',
        metal_FF=test_molecule.metal_ff,
    )

    # Assign the force field.
    if test_molecule.has_metal and test_molecule.metal_ff is None:
        with pytest.raises(ExpectedMetal):
            gulp_opt.assign_FF(test_molecule.molecule)
    else:
        gulp_opt.assign_FF(test_molecule.molecule)

        assert (
            len(gulp_opt.atom_labels) == len(test_molecule.atom_types)
        )
        for aid in gulp_opt.atom_labels:
            print(aid)
            expected_type = test_molecule.atom_types[aid]
            test = gulp_opt.atom_labels[aid][0]
            assert expected_type == test
